// dmx.c  // ***************************************************************
//
// DMX related files for
// simple DMX512 controller for STM32G431KB (32 pin LQFP package)
// based on USB MIDISTREAM and CDC-ABSTRACT CONTROL MODEL
//
// Author and (c)2020-21 Wolfgang Schemmert. 
// The copyright claim and GPL-3 licence is restricted to this file
// and the depending files "main.c","main.h","usb-fs-431.c", "userinterface.c"
// and "globaldefines.h".
// Other depending material is published and copyrighted
// by STMicroelectronics and Segger GmbH/Rowley Associates Ltd
//
// Contact: <www.midi-and-more.de>
// A detailled hardware construction and operation manual 
// is provided at this website
//
// ****************************************************************************
//      This program is free software: you can redistribute it and/or modify
//      it under the terms of the GNU General Public License as published by
//      the Free Software Foundation, either version 3 of the License, or
//      (at your option) any later version.
//
//      This program is distributed in the hope that it will be useful,
//      but WITHOUT ANY WARRANTY; without even the implied warranty of
//      MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//      GNU General Public License for more details.
//
//      You should have received a copy of the GNU General Public License
//      along with this program.  If not, see <http://www.gnu.org/licenses/>.
//
//      Licence: GPL-3
//
// ****************************************************************************

#include "stm32g431xx.h"
#include "globaldefines.h"

// ************************************

//main.h
extern uint32_t gSYSCOUNT, gLEDCOUNT, gTIMEOUT, gDELAYCOUNT;
extern uint32_t gVid, gPid, gChannel, gLedBlink;
extern void popDmxFlash(uint32_t);
extern void msDelay(uint32_t);
extern void sendUsbByte(uint8_t);
extern void sendUsbString(const uint8_t *);
extern void sendUsbNibble(uint8_t);
extern void sendUsbX8(uint8_t);
extern void sendUsbX16(uint16_t);
extern void sendUsbX32(uint32_t);
extern void sendUsbN32(uint32_t);   //sendS32() without trailing SPACE
extern void sendUsbS32(uint32_t);
extern uint16_t pullUsbRxBuf(void);

//usb-fs.c
extern void sendMidiMsg(uint8_t, uint8_t, uint8_t, uint8_t);

//userinterface.c
extern uint32_t parseDeciNum(uint32_t);
extern const uint8_t gcFadeStr[];
extern const uint8_t gcCycleStr[];

// ************************************

uint8_t gdMINIDMX_ON;  //gdNoZeroVelocity,

uint8_t gdDMXBUF[DMX_BUFLEN+2];     //actual DMX transmit buffer
            //reserve dummies at index 0 and index 513-519 for practical reasons
uint8_t gdFINALMEM[DMX_BUFLEN+2];   //final level of fade process
uint8_t gdSTEPMEM[DMX_BUFLEN+2];    //used for control of fade timing
uint8_t gdSUBMEM[DMX_BUFLEN+2];
uint8_t gdRESIMEM[DMX_BUFLEN+2];
//uint8_t gdTempBuf[2048];
uint32_t gdDmxAddr, gdDmxTxIndex;
uint32_t gdFaderBusy, gdMiniDmxOn;
uint32_t gdDmxMax, gdFadeTime=0;
uint32_t gdGlobalHue, gdGlobalSat, gdGlobalLum;
uint32_t gdChaseStart, gdChaseLen, gdChaseTime, gdChaseStep, gdChaseCount;
//gdChaseStart is first preset of chaser cycle
//gdChaseLen is number of presets called every chaser cycle
//gdChaseTime is duration of each chaser step in 1/10 seconds
//gdChaseStep is offset in cycle (shown preset = gdChaserStart+gdChaseStep)
//gdChaseCount is countdown of gdChaserTime in 1/10 second intervals

void startDmxTx(void);
void DmxHeader(void);
void wait32us(void);
void sendDmxByte(void);
void MiniDMX_Interpreter(void);
void MiniDMX_ACK(void);
void setDmxAddress(void);
void setDmxLevel(uint8_t);
void setNextDmxLevel(uint8_t);
void hsb2RGB(void);
void setEqual(uint8_t);
void fade2Preset(void);
void startFader(uint16_t);
void fadeProcess(void);
void nextChaserStep(void);
void msgDmxState(uint32_t);
void readDmxBuf(uint32_t, uint32_t);
void sendAddrLevel(void);
void sendLevel(uint16_t);

// ************************************

// DMX handling
//***************

void startDmxTx(void) // ****************************************************
// MINI:USART1-TxD is PortB6,  DMX-Break (Open Drain Pull Down) is PB5
// **************************************************************************
{
uint32_t k;

	RCC->APB2ENR |=  RCC_APB2ENR_USART1EN ; // USART1 clock enable

	GPIOB->MODER &= ~(GPIO_MODER_MODER6_0 ); //Set PB66 to alternate function
	GPIOB->MODER |= GPIO_MODER_MODER6_1 ;  // 10=Alternate function mode
	//GPIOA->OTYPER is push-pull by default
	GPIOB->OSPEEDR &= ~(GPIO_OSPEEDR_OSPEED6_1);
	GPIOB->OSPEEDR |= GPIO_OSPEEDR_OSPEED6_0;  //Medium Speed

// PB5 is Open Drain and pulls DMX OUT to SPACE during DMX Break=Reset
	//configure PA10 as open drain output, zero drive current
	GPIOB->OTYPER |= GPIO_OTYPER_OT_5;
	GPIOB->BSRR = GPIO_BSRR_BS_5;			//Open Drain High= transistor OFF!
	GPIOB->MODER &= ~GPIO_MODER_MODER5_1;
	GPIOB->MODER |= GPIO_MODER_MODER5_0;//PortA0 general purpose output mode
	GPIOB->OSPEEDR &= ~GPIO_OSPEEDR_OSPEED5_1;
	GPIOB->OSPEEDR |= GPIO_OSPEEDR_OSPEED5_0;  //Medium Speed

//USART1 is AF7 on PortA
	GPIOB->AFR[0] &= 0xF0FFFFFF;	//PB6 is TX
	GPIOB->AFR[0] |= 0x07000000;

#ifdef CPU_CLOCK144
	USART1->BRR = 288; //250,000 Baud at 72MHz APB Clock (288=72000000/25000)
#else
	USART1->BRR = 192; //250,000 Baud at 48MHz APB Clock (192=48000000/25000)
#endif
	USART1->CR2 = USART_CR2_STOP_1;	//=2 Stopbits
	USART1->CR3 = 0;				//default:no Handshake
	USART1->CR1 = USART_CR1_UE | USART_CR1_TE;
									//TX only, 8Bit, no Parity, no Interrupts
	for (k=0;k < (DMX_BUFLEN+1);k++)  {
		gdDMXBUF[k] = 0;
		gdFINALMEM[k] = 0;
		gdSTEPMEM[k] = 0;
		gdSUBMEM[k] = 0;
		gdRESIMEM[k] = 0;
	}
	gdDmxAddr = 0;
//no interrupt is used
}


void DmxHeader(void)  // *****************************************************
// ***************************************************************************
{
uint8_t waitcount;

// Mark before Break min 0 max 1sec Includes transmission time of last DMX byte
	waitcount = 12;					//tested ca. 350 us
	while (waitcount)  {
		wait32us();
		waitcount-- ;
	}

// Space for Break = DMX loop Reset "typically" ca 176 us
// switch Open Drain output
	GPIOB->BSRR = GPIO_BSRR_BR_5;	//Open Drain Low= transistor ON !
	waitcount = 8;					//tested ca 280us
	while (waitcount)  {
		wait32us();
		waitcount-- ;
	}
// release Open Drain output
    GPIOB->BSRR = GPIO_BSRR_BS_5;	//Open Drain High= transistor OFF !
// Mark after Break min 12 us
	wait32us();						//tested ca 32us

// send DMX Startbyte = 0
	while ((USART1->ISR & USART_ISR_TXE) == 0) IWDG->KR = 0x0000AAAA;
	//Emergency loop: TXE normally should be already == 1;feed watchdog.
	USART1->TDR = 0;	//set DMX StartBit
// the TXE bit is cleared automatically when TDR is written
	gdDmxTxIndex = 0;               //index 0 of DMX buffer is User DMX addr=1
}


void wait32us(void)   // *******************************************************
// blocking delay ca 32us
// *************************************************************************
{
uint32_t countdown;

#ifdef CPU_CLOCK144
	countdown = 600;
#else
	countdown = 400;
#endif
	while (countdown)   {
		IWDG->KR = 0x0000AAAA;  //timeout ca. 1-2 seconds
		countdown-- ;
	}
}


void MiniDMX_Interpreter(void)  // ******************************************
// **************************************************************************
{
uint8_t tempBuf[DMX_BUFLEN];
uint16_t recvword, index, packetLen;
uint8_t recvbyte;

    gTIMEOUT = 100;

    while (1)  {
        if (gTIMEOUT == 0) {        //seems to be non-MiniDMX input
            gdMiniDmxOn = 0;        //MiniDmx regarded as inactive
            gLedBlink = 0;
            return;
        }
        recvword = pullUsbRxBuf();
        if (recvword)  {
            recvbyte = (uint8_t)(recvword & 0x00FF);

            if (recvbyte == 0xA2) {
                packetLen = 512;
                break;      // OK, start next stage
            }
            else if (recvbyte == 0xA1) {
                packetLen = 256;
                break;      // OK, start next stage
            }
            else if (recvbyte == 0xA0) {
                packetLen = 96;
                break;      // OK (but NOT TESTED HERE!), start next stage
            }
            else  {                 //unvalid packet specifier: error
                MiniDMX_ACK();      //send ACK even in case of error
                return;             //but stay in MiniDMX mode
            }
        }
    }

    gdMiniDmxOn = 0xFF;         //MiniDmx regarded as active

//valid MiniDMX packet opened
    if (gLedBlink==0) gLedBlink = 1;    //don't restart, else not blinking

    index = 0;  //starts from 0 LIKE gDmxTxIndex !!

    while (1)  {
        gTIMEOUT = 60;           //60ms timeout per byte

        while (1)  {
            recvword = pullUsbRxBuf();
            if (recvword)  {
                recvbyte = (uint8_t)(recvword & 0x00FF);
                break;          //next byte received successfully
            }

            if(gTIMEOUT == 0)  {
                MiniDMX_ACK();  //send ACK even in case of error
                return;         //but stay in MiniDMX mode
            }

        }
//store bytes temporarily not to disturb DMX transmission in case of error
        tempBuf[index] = recvbyte;
//packet is complete when number of DMX bytes specified by protocol
//plus 1 termination byte (0xA5) received
        if (index >= packetLen) break;
        index++ ;
    }
//byte transfer complete --> next stage:
    if (recvbyte != 0xA5) { //check for correct packet termination
        MiniDMX_ACK();      //send ACK even in case of error
        return;             //but stay in MiniDMX mode
    }
//packet received successfully, store into DMX transmit buffer
    for (index=0; index < packetLen; index++) {
        gdDMXBUF[index] = tempBuf[index];
        gdFINALMEM[index] = tempBuf[index];       //terminate any fading,too
    }
    MiniDMX_ACK();
}


void MiniDMX_ACK(void)  // **************************************************
// **************************************************************************
{
    sendUsbByte(0x5A);
    sendUsbByte(0xC1);  //send ACK even in case of error
    sendUsbByte(0xA5);  //then better get next actual one
}


void setDmxAddress(void)  // ************************************************
// only get and store gDmxAddr as write pointer to DMX transmit buffer
// **************************************************************************
{
uint32_t number;

    number = parseDeciNum(512);
    if ((number > 512) || (number == 0)) {
        sendUsbByte('?');
        return;
    }
    else gdDmxAddr = number - 1;
}


void setDmxLevel(uint8_t level)  // *****************************************
// **************************************************************************
{
        if (gdFadeTime == 0)  {
        gdFINALMEM[gdDmxAddr] = level;
        gdDMXBUF[gdDmxAddr] = level;
    }
    else  {
        gdFINALMEM[gdDmxAddr] = level;
        startFader(gdDmxAddr);
    }
}


void setNextDmxLevel(uint8_t level)  // *************************************
// **************************************************************************
{
    if (gdDmxAddr == 511)  {
        sendUsbByte('?');
        return;
    }
    gdDmxAddr++ ;
    setDmxLevel(level);
}


void hsb2RGB(void)  // ******************************************************
// **************************************************************************
{
uint32_t red, green, blue, param, param2;
uint32_t address;

// update Hue
	param = gdGlobalHue;
	if (param < 43)  {
		param = param << 1;
		red = 255 - param;		//red decreases down to 171
		green = param << 1;		//param*4:green increases from 0 to 168
		blue = 0;
	}
	else if (param < 85)  {
		param -= 43;					//range 0...42
		param2 = param >> 3;				//range 0...5	
		green = 171 +  (param << 1) ;		//increases from 171 to 253
		param = (param << 2) + 2;
		red = 172 - param - param2;		//decreaes from 170 to 1
		blue= 0;
	}
	else if  (param < 128)  {
		red = 0;
		param -= 85;					//range 0...42
		param = param << 1;
		green = 255 - param;			//green decreases down to 171	
		blue = param << 1;				//param*4:blue increases from 0 to 168
	}
	else if (param < 170)  {
		red = 0;
		param -= 128;					//range 0...41
		param2 = param >> 3;				//range 0...5	
		blue = 171 +  (param << 1) ;		//blie increases from 171 to 253
		param = (param << 2) + 2;
		green = 172 - param - param2;	//green decreaes from 170 to 1
	}
	else if (param < 213)  {
		param -= 170;					//range 0...42
		param = param << 1;
		blue = 255 - param;				//blue decreases down to 171	
		red = param << 1;				//param*4:red decreases from 0 to 168
		green = 0;
	}
	else  {
		param -= 213;					//range 0...42
		param = param << 1;
		red = 171 + param;				//red increases from 171 to 255	
		blue = 170 -(param << 1);		//param*4:blue decreases from 168 to 0
		green = 0;
	}

//update Saturation
	param = gdGlobalSat;
	if (param < 100)  {
		param = 100 - param;
		param2 = 0x000000FF - red;
		param2 = (param2 * param) / 100;
		red += param2;
		param2 = 0x000000FF - green;
		param2 = (param2 * param) / 100;
		green += param2;
		param2 = 0x000000FF - blue;
		param2 = (param2 * param) / 100;
		blue += param2;
	}
// if (param ==100) 100% Saturation = no change of Hue

//update Luminance
	param = gdGlobalLum;
	if  (param < 100)  {
		red = (red * param) / 100;
		green = (green * param) / 100;
		blue = (blue * param) / 100;
	}
//param = 100 means no decrease of actual red, green. blue
//final transfer to addressed DXM channels
	address = gdDmxAddr;
	setDmxLevel((uint8_t) red);
	setNextDmxLevel((uint8_t) green);
	setNextDmxLevel((uint8_t) blue);
	gdDmxAddr = address;

	sendUsbByte(' ');
	sendUsbS32(red);
	sendUsbS32(green);
	sendUsbS32(blue);
	sendUsbByte(CRR);
}	


void setEqual(uint8_t number)  // *******************************************
// **************************************************************************
{
uint8_t level;
uint16_t address,k;

    if (number == 0) return;
    level = gdFINALMEM[gdDmxAddr];
       address = gdDmxAddr;
    for (k=0;k<number;k++)  {
        gdDmxAddr++;
        setDmxLevel(level);
        if (gdDmxAddr == 511)  break;
    }
    gdDmxAddr = address;
}


void fade2Preset(void)  // **************************************************
// **************************************************************************
{
int16_t k;

    for (k=0;k<DMX_BUFLEN;k++)  {
        if (gdDMXBUF[k] == gdFINALMEM[k]) continue;
        if (gdFadeTime == 0)  gdDMXBUF[k] = gdFINALMEM[k];
        else startFader(k);
    }
}


void startFader(uint16_t address)  // ***************************************
// **************************************************************************
{
uint32_t dividend, quotient;

    if (gdFadeTime == 0) return;//precaution,usually inhibited by program flow
    if (gdDMXBUF[address] == gdFINALMEM[address]) return;  //all-ready

    dividend = gdDMXBUF[address] - gdFINALMEM[address];
//calculate absolute difference,
    if (gdDMXBUF[address] > gdFINALMEM[address])
                            dividend = gdDMXBUF[address] - gdFINALMEM[address];
    else dividend =  gdFINALMEM[address] - gdDMXBUF[address];

    quotient = (dividend << 8) / gdFadeTime;
                                    //divisor != 0 guaranteed by program flow
    gdSTEPMEM[address] =(uint8_t)(quotient >> 8);    //HighByte
//gdSTEMPMEM contains the number of full gdDMXBUF steps each /50s  (fast fade)
    gdSUBMEM[address] =(uint8_t)(quotient & 0x00FF); //LowByte
//because the the dividend was multiplied by 256, SUBMEM is a kind of remainder
//this is added to RESIMEM (residual of full step), ie. when RESIMEM overflows,
//a full extra +/-step (depends on fade direction) for DMXBUF gets inserted
//(slow fade), execution of this see fadeProcess().
    gdRESIMEM[address] = 0;
    gdFaderBusy = 1;
}


void fadeProcess(void)  // **************************************************
//this is the "ballistic" algorithm (target is computed at start time), with
//very small CPU overhead at runtime. It is less precise for long time fades
//with small differences between actual and final level.
//an alternative would be the "cruise missile" algorithm, which calculates the
//difference between actual and target every /50s and is more precise,
//but takes more CPU time (a division per DMX channel update).
// **************************************************************************
{
uint16_t k, residual;
int16_t actual, temp;
uint8_t step;

    gdFaderBusy = 0;                //temporary,check if further fade is active

    for (k=0;k<DMX_BUFLEN;k++)  {
        if (gdDMXBUF[k] == gdFINALMEM[k]) continue;
        gdFaderBusy = 0xFF;                 //trigger next fadeProcess()
        actual = (uint16_t)gdDMXBUF[k];
        step = gdSTEPMEM[k];
        residual = (uint16_t)gdRESIMEM[k];
        residual += gdSUBMEM[k];
        gdRESIMEM[k] = (uint8_t)(residual & 0x00FF);
        if (residual >= 0x0100) step++; //handle overflow without side effect
        if (actual < (uint16_t)gdFINALMEM[k])  {     //fade UP
            temp = (int16_t)(gdDMXBUF[k]) + step;
            if (temp > gdFINALMEM[k]) actual = gdFINALMEM[k];
            else actual += step;
        }
        else  {                                 //fade DOWN
            temp = (int16_t)(gdDMXBUF[k]) - step;
            if (temp < gdFINALMEM[k]) actual = gdFINALMEM[k];
            else  actual -= step;
        }
        gdDMXBUF[k] = (uint8_t)actual;
    }
}


void nextChaserStep(void)  // ***********************************************
// **************************************************************************
{
uint32_t offset;

	gdChaseCount = gdChaseTime;	//start new countdown cycle
	offset = gdChaseStart + gdChaseStep;
	if (offset > MAXPRESET)  {
		offset = gdChaseStart;
		gdChaseStep = 0;
	}
	popDmxFlash(offset);
	gdChaseStep ++;
	if (gdChaseStep == gdChaseLen) gdChaseStep = 0;
}


void msgDmxState(uint32_t mode)  // *****************************************
// **************************************************************************
{
uint8_t timeH;
const uint8_t addressStr [] = "DMX channel:";
const uint8_t channelStr [] = "    channel:";
const uint8_t finalStr [] = "Final:";
const uint8_t chaserStr [] = "Chaser Start Scene:";

	if (mode) sendUsbByte(CRR);
	sendUsbString(addressStr);
	sendUsbS32(gdDmxAddr+1);
	sendUsbByte('T');
	sendUsbByte('x');
	sendUsbByte(':');
	sendUsbS32(gdDMXBUF[gdDmxAddr]);
	sendUsbString(finalStr);
	sendUsbS32(gdFINALMEM[gdDmxAddr]);
	sendUsbByte(' ');

	sendUsbByte(0x28);		//'('
	sendUsbString(gcFadeStr);
	timeH = gdFadeTime/50;
	sendUsbN32(timeH);
	sendUsbByte('.');
	sendUsbN32((gdFadeTime/5) - (timeH*10));
	sendUsbByte('s');
	sendUsbByte(' ');
	sendUsbByte(' ');
	sendUsbString(gcCycleStr);
	sendUsbN32(gdDmxMax);
	sendUsbByte(0x29);		//')'
	if (mode)  {
		sendUsbByte(CRR);
		if (gdDmxAddr < 510)  {
			sendUsbString(channelStr);
			sendUsbS32(gdDmxAddr+2);
			sendUsbByte('T');
			sendUsbByte('x');
			sendUsbByte(':');
			sendUsbS32(gdDMXBUF[gdDmxAddr+1]);
			sendUsbString(finalStr);
			sendUsbS32(gdFINALMEM[gdDmxAddr+1]);
			sendUsbByte(CRR);
			sendUsbString(channelStr);
			sendUsbS32(gdDmxAddr+3);
			sendUsbByte('T');
			sendUsbByte('x');
			sendUsbByte(':');
			sendUsbS32(gdDMXBUF[gdDmxAddr+2]);
			sendUsbString(finalStr);
			sendUsbS32(gdFINALMEM[gdDmxAddr+2]);
			sendUsbByte(CRR);
		}
	}
	else 	sendUsbByte(' ');
	sendUsbString(chaserStr);
	sendUsbS32(gdChaseStart);
	sendUsbByte('L');
	sendUsbByte('e');
	sendUsbByte('n');
	sendUsbByte(':');
	sendUsbS32(gdChaseLen);
	sendUsbByte('S');
	sendUsbByte('t');
	sendUsbByte('e');
	sendUsbByte('p');
	sendUsbByte(':');
	sendUsbN32(gdChaseTime/10);
	sendUsbByte('.');
	sendUsbNibble(gdChaseTime%10);
	sendUsbByte('s');
}


void readDmxBuf (uint32_t mode, uint32_t number)  // ************************
// **************************************************************************
{
uint16_t address;

	if (mode)  {
			sendUsbByte(' ');
			sendUsbByte('f');
			sendUsbByte('r');
			sendUsbByte('o');
			sendUsbByte('m');
			sendUsbByte(' ');
	}

    sendUsbByte('S');
    sendUsbN32(gdDmxAddr+1);
    sendUsbByte(':');
    sendUsbByte('V');
    sendUsbByte('=');
    address = gdDmxAddr;
    sendUsbN32((int16_t)gdDMXBUF[address]);
    address++ ;
    number -- ;
    while (1)  {
        if ((address == DMX_BUFLEN) || (number== 0)) break;
        sendUsbByte(',');
        sendUsbN32((int16_t)gdDMXBUF[address]);
        address++;
        number--;
    }
}


void sendAddrLevel(void)  // *************************************************
//first return actually addressed DMX channel as Control Change 0x50...0x53
//then return this actual DMX level as Control Change 0x54 or 0x55
// **************************************************************************
{
uint8_t data1, data2;

    if (gdDmxAddr == 511)  {         	//gets responded as DMX channel no.0
        data1 = 0x50;
        data2 = 0;
    }
    else  {
       data1 = (uint8_t)((gdDmxAddr + 1)>>7) + 0x50;
							//internally gdDmxAddr is counted from 0 to 511
       data2 =(uint8_t)((gdDmxAddr + 1) & 0x007F);
    }

//"sendUsbByte()" calls "enCodeMidi()":
    sendUsbByte(0xB0 | (uint8_t)(gChannel & 0x00FF)); 	//Status
    sendUsbByte(data1);
    sendUsbByte(data2);
//append the actual Fade Time
//"sendUsbByte()" calls "enCodeMidi()":
    sendUsbByte(0xB0 | (uint8_t)(gChannel & 0x00FF)); 	//Status
    sendUsbByte(1);                                   	//Data1
    sendUsbByte(gdFadeTime / 5);                      	//Data2

    sendLevel(gdDmxAddr);
}


void sendLevel(uint16_t address)  // ****************************************
//return DMX level of the addressed DMX channel as Control Change 0x54 or 0x55
// **************************************************************************
{
//"sendUsbByte()" calls "enCodeMidi()":
    sendUsbByte(0xB0 | (uint8_t)(gChannel & 0x000000FF));	//Status
    sendUsbByte((uint8_t)(gdDMXBUF[address] & 1) + 0x54); 	//Data1
    sendUsbByte((uint8_t)(gdDMXBUF[address] >>1));      	//Data2
}

// ##########################################################################

